from collections.abc import MutableMapping
from .exceptions import EncoderException
from .serialize import serializers


class BaseStore(MutableMapping):
    """
    The Store represents storage for any key-value pair.

    This is an abstract class and should not be used directly.
    """

    __slots__ = ('_type', '_encode', '_decode')

    def __init__(self, serialize=None, dump_load=tuple(), value_type=bytes):
        self._type = value_type
        if serialize and serialize in serializers:
            self._encode = serializers[serialize]['encode']
            self._decode = serializers[serialize]['decode']
        elif dump_load and len(dump_load) == 2:
            self._encode, self._decode = dump_load
        else:
            raise EncoderException('The store needs an encoder and a decoder')

    def _enc_key(self, k):
        if isinstance(k, str):
            return k.encode('utf8')
        return k

    def get(self, key):
        raise NotImplementedError

    def put(self, key, value):
        raise NotImplementedError

    def delete(self, key):
        raise NotImplementedError

    def close(self):
        raise NotImplementedError

    def destroy(self, yes_im_sure=False):
        raise NotImplementedError

    def setdefault(self, key, default=None):
        """
        If key is in the dictionary, return its value.
        If not, insert key with a value of default and return default.
        """
        if key in self:
            return self[key]
        self[key] = default
        return default

    def deep_add(self, key, value):
        """
        Add a value in the deep structure found at key.
        The deep structure can be a list, or a set.
        """
        data = self._type(self.get(key, []))
        if hasattr(data, 'add'):
            data.add(value)
        elif hasattr(data, 'append'):
            data.append(value)
        else:
            raise TypeError('Cannot add value in a "{}"'.format(type(data)))
        self[key] = data

    def deep_remove(self, key, value):
        """
        Remove a value from the deep structure found at key.
        The deep structure can be a list, or a set.
        """
        data = self._type(self.get(key, []))
        if hasattr(data, 'discard'):
            data.discard(value)
        elif hasattr(data, 'remove'):
            data.remove(value)
        else:
            raise TypeError('Cannot del value from a "{}"'.format(type(data)))
        self[key] = data



"""
Functions that serialize Python objects.
Pickle and JSON are builtin libraries.
CBOR2 and MsgPack are available with PIP.
Other than the normal objects, available in all languages,
the `tuple`, `set` and `frozenset` are also supported.
"""

import pickle
from io import BytesIO
from .util import ensure_bytes

try:
    # Ultra fast JSON encoder and decoder
    import ujson as json
except ModuleNotFoundError:
    try:
        # Second fast JSON encoder and decoder
        import simplejson as json
    except ModuleNotFoundError:
        print('For better performance using JSON, install either UJSON or simplejson')
        import json

try:
    import cbor2
    from cbor2.types import CBORTag
except ModuleNotFoundError:
    cbor2 = None
    print('CBOR can be installed at PyPi.python.org/pypi/cbor2')

try:
    import msgpack
except ModuleNotFoundError:
    msgpack = None
    print('MessagePack can be installed at PyPi.python.org/pypi/msgpack')

TUP_FLAG = b'__(,)__'
SET_FLAG = b'__{,}__'
F_SET_FLAG = b'__f{}__'  # frozenset

TUP_CBOR = 38


def noop(data):
    """
    No-op "encoder" returns the object exactly as it is.
    """
    return data


def encode_pickle(data):
    """
    Serialize an object using Pickle's highest protocol.
    Pickle doesn't need any help dealing with custom Python structures.
    """
    return pickle.dumps(data, protocol=pickle.HIGHEST_PROTOCOL, fix_imports=False)


def decode_pickle(data):
    """
    Restore original object using Pickle.
    """
    return pickle.loads(data, fix_imports=False)


def _convert_python_obj(data):
    """
    Convert tuple, set and frozenset into list
    and insert a special flag in the first position,
    to recognize how the list should be converted back.
    Used by JSON and MsgPack.
    """
    if isinstance(data, tuple):
        return [TUP_FLAG] + list(data)
    if isinstance(data, set):
        return [SET_FLAG] + list(data)
    if isinstance(data, frozenset):
        return [F_SET_FLAG] + list(data)
    return data


def _restore_python_obj(data):
    """
    Convert flagged list back to a Python object.
    Used by JSON and MsgPack.
    """
    if isinstance(data, (bytes, str)):
        return ensure_bytes(data)
    if isinstance(data, list):
        if data[0] == TUP_FLAG:
            return tuple(data[1:])
        if data[0] == SET_FLAG:
            return set(data[1:])
        if data[0] == F_SET_FLAG:
            return frozenset(data[1:])
    return data


def encode_json(data):
    """
    Serialize an object using JSON.
    Convert special Python iterable into list.
    """
    data = _convert_python_obj(data)
    return ensure_bytes(json.dumps(data))


def decode_json(data):
    """
    Restore original object using JSON.
    JSON doesn't know how to deal with bytes and it needs help.
    """
    data = json.loads(data)
    if isinstance(data, (bool, int, float)):
        return data
    if isinstance(data, (bytes, str)):
        return ensure_bytes(data)
    if isinstance(data, dict):
        return {_restore_python_obj(k): _restore_python_obj(v) for k, v in data.items()}
    return _restore_python_obj([_restore_python_obj(e) for e in data])


def _cbor_encoder(encoder, data):
    """
    Hook to convert tuples and sets into lists.
    The list of CBOR tags:
    http://cbor2.readthedocs.io/en/latest/usage.html#tag-support
    """
    if isinstance(data, tuple):
        encoder.encode(CBORTag(TUP_CBOR, list(data)))


def _cbor_decoder(decoder, tag, _):
    """
    Hook to convert lists into their original objects.
    """
    if tag.tag == TUP_CBOR:
        return tuple(tag.value)


def encode_cbor(data):
    """
    Serialize an object using CBOR.
    The CBOR Encoder is needed to revert the default
    behaviour of treating tuples like normal lists.
    """
    mem_file = BytesIO()
    encoder = cbor2.CBOREncoder(mem_file, default=_cbor_encoder)
    del encoder._encoders[tuple]  # Enable the default callback
    encoder.encode(data)
    return mem_file.getvalue()


def decode_cbor(data):
    """
    Restore original object using CBOR.
    """
    return cbor2.loads(data, tag_hook=_cbor_decoder)


def encode_msgpack(data):
    """
    Serialize an object using MsgPack.
    """
    return msgpack.dumps(data, use_bin_type=True, strict_types=True, default=_convert_python_obj)


def decode_msgpack(data):
    """
    Restore original object using MsgPack.
    """
    return msgpack.loads(data, list_hook=_restore_python_obj)


serializers = {
    'noop': {
        'encode': noop,
        'decode': noop
    },
    'pickle': {
        'encode': encode_pickle,
        'decode': decode_pickle
    },
    'json': {
        'encode': encode_json,
        'decode': decode_json
    }
}

if cbor2:
    serializers['cbor'] = {'encode': encode_cbor, 'decode': decode_cbor}

if msgpack:
    serializers['msgpack'] = {'encode': encode_msgpack, 'decode': decode_msgpack}



import itertools
import contextlib
from .base import BaseStore


class MemoryStore(BaseStore):
    """
    Pure memory store.
    Inspired from builtin collections.UserDict.
    """

    __slots__ = ('db',)

    def __init__(self,
                 *arg,
                 serialize='noop',
                 dump_load=tuple(),
                 value_type=bytes,
                 iterable=tuple(),
                 kwargs={}):
        super().__init__(serialize=serialize, dump_load=dump_load, value_type=value_type)
        self.db = {}
        if iterable or kwargs:
            self._populate(iterable, **kwargs)

    def _populate(self, iterable=tuple(), **kwargs):
        with contextlib.suppress(AttributeError):
            iterable = iterable.items()
        for key, value in itertools.chain(iterable, kwargs.items()):
            self.db[key] = self._encode(value)

    def get(self, key, default=None):
        encoded_value = self.db.get(self._enc_key(key))
        return self._decode(encoded_value) if encoded_value else default

    def put(self, key, value, overwrite=True):
        enc_key = self._enc_key(key)
        if not overwrite and self.db.get(enc_key):
            return
        self.db[enc_key] = self._encode(value)

    def delete(self, key):
        del self.db[self._enc_key(key)]

    __getitem__ = get

    __setitem__ = put

    __delitem__ = delete

    def __contains__(self, key):
        return self._enc_key(key) in self.db

    def __len__(self):
        return len(self.db)

    def __iter__(self):
        return iter(self.db)

    def __repr__(self):
        return self.__class__.__name__ + repr(self.db)

    def keys(self):
        return list(self.db.keys())

    def values(self):
        return [self._decode(v) for v in self.db.values()]

    def items(self):
        items_list = []
        for key, value in self.db.items():
            items_list.append((key, self._decode(value)))
        return items_list

    def update(self, iterable=tuple(), **kwargs):
        self._populate(iterable, **kwargs)

    def close(self):
        pass

    def clear(self):
        self.db.clear()

    def destroy(self, yes_im_sure=False):
        if yes_im_sure:
            self.db.clear()




def ensure_bytes(text):
    if isinstance(text, bytes):
        return text
    if isinstance(text, str):
        return text.encode('utf')
    raise TypeError('Cannot convert {} type into bytes'.format(type(text)))



# flake8: noqa

from .memory import MemoryStore
from .lmdb import LmdbStore
from .serialize import serializers
from .exceptions import EncoderException


def stone(name, store=MemoryStore, serialize='pickle', dump_load=tuple(), value_type=bytes, options={}):
    """
    Convenience wrapper for returning a Store instance.
    """
    if serialize and serialize not in serializers:
        raise EncoderException('Invalid serializer name "{}"'.format(serialize))
    return store(name, serialize=serialize, dump_load=dump_load, value_type=value_type, **options)



import shutil
import itertools
import contextlib
from .base import BaseStore

try:
    import lmdb
except ModuleNotFoundError:
    print('LMDB store requires PyPi.python.org/pypi/lmdb')

LMDB_ENVIRONMENT = {'max_dbs': 9, 'map_size': 8e12}


class LmdbStore(BaseStore):
    """
    LMDB ‘Lightning’ Database container, compatible with Python dicts.
    Keys and values MUST be byte strings.
    """

    __slots__ = ('db', 'table', '_name')

    def __init__(self,
                 name,
                 table=None,
                 serialize='json',
                 dump_load=tuple(),
                 value_type=bytes,
                 iterable=tuple(),
                 database={},
                 kwargs={}):
        super().__init__(serialize=serialize, dump_load=dump_load, value_type=value_type)
        self._name = name
        self.db = lmdb.open(self._name, **{**LMDB_ENVIRONMENT, **database})
        self.table = self.db.open_db(table)
        if iterable or kwargs:
            self._populate(iterable, **kwargs)

    def close(self):
        self.db.close()

    def _populate(self, iterable=tuple(), **kwargs):
        with contextlib.suppress(AttributeError):
            iterable = iterable.items()
        with self.db.begin(write=True, db=self.table) as txn:
            for key, value in itertools.chain(iterable, kwargs.items()):
                txn.put(key, self._encode(value), dupdata=False)

    # def put_bytes(self, key, raw, dupdata=False, overwrite=False):
    #     """ Inject raw bytes """
    #     if not isinstance(raw, bytes):
    #         raise TypeError('Raw data must be Bytes')
    #     with self.db.begin(write=True, db=self.table) as txn:
    #         txn.put(key, raw, dupdata=dupdata, overwrite=overwrite)

    def get(self, key, default=None):
        with self.db.begin(db=self.table) as txn:
            encoded_value = txn.get(self._enc_key(key), None)
            return self._decode(encoded_value) if encoded_value else default

    def put(self, key, value, overwrite=False):
        enc_key = self._enc_key(key)
        with self.db.begin(write=True, db=self.table) as txn:
            txn.put(enc_key, self._encode(value), overwrite=overwrite, dupdata=False)

    def delete(self, key):
        with self.db.begin(write=True, db=self.table) as txn:
            return txn.delete(self._enc_key(key))

    def __getitem__(self, key):
        with self.db.begin(db=self.table) as txn:
            encoded_value = txn.get(self._enc_key(key))
            return self._decode(encoded_value)

    def __setitem__(self, key, value):
        with self.db.begin(write=True, db=self.table) as txn:
            txn.put(self._enc_key(key), self._encode(value), dupdata=False)

    __delitem__ = delete

    def __contains__(self, key):
        enc_key = self._enc_key(key)
        with self.db.begin(db=self.table) as txn:
            return bool(txn.get(enc_key))

    def __len__(self):
        with self.db.begin(db=self.table) as txn:
            return txn.stat()['entries']

    def __iter__(self):
        with self.db.begin(db=self.table) as txn:
            yield from txn.cursor().iternext(keys=True, values=False)

    def __repr__(self):
        items = dict(self.items())
        return self.__class__.__name__ + repr(items)

    def keys(self):
        keys_list = []
        with self.db.begin(db=self.table) as txn:
            for key in txn.cursor().iternext(keys=True, values=False):
                keys_list.append(key)
        return keys_list

    def values(self):
        vals_list = []
        with self.db.begin(db=self.table) as txn:
            for value in txn.cursor().iternext(keys=False, values=True):
                vals_list.append(self._decode(value))
        return vals_list

    def items(self):
        items_list = []
        with self.db.begin(db=self.table) as txn:
            for key, value in txn.cursor().iternext(keys=True, values=True):
                items_list.append((key, self._decode(value)))
        return items_list

    def update(self, iterable=tuple(), **kwargs):
        self._populate(iterable, **kwargs)

    def clear(self):
        with self.db.begin(write=True) as txn:
            txn.drop(self.table, delete=False)

    def destroy(self, yes_im_sure=False):
        if yes_im_sure:
            self.close()
            shutil.rmtree(self._name)
